// PortProxy 99 by Steven Lawrance
// (c) 1999 Steven Lawrance

#include <winsock2.h>
#include <windows.h>
#include "ip.h"
#include "udp.h"
#include "checksum.h"
#include "resource.h"

#define SERVICE_NAME "PortProxy"
#define MYWM_NOTIFYICON WM_USER + 50

// Variable for exiting
BOOL bExit;
HINSTANCE ghInstance;
BOOL bAsking;

// Debug (non-service/console) mode flag
//BOOL bDebugMode;

// Critical section for main loop
//CRITICAL_SECTION csMainLoop;

// Readiness states
#define STATUS_READY    0
#define STATUS_STARTING 1
#define STATUS_CLOSING  2

// Registration structure
typedef struct _REGISTRATION_INFO
{
  SOCKET    hSocket;
  BOOL      bExit;
  WORD      wPort;
  char      szInternalAddress[16];   // "www.xxx.yyy.zzz" format
  WORD      wInternalPort;
  HANDLE    hThread;
  struct sockaddr_in saInternal;
  LPVOID    next;
  DWORD     dwRegId;
  DWORD		dwType;
} REGISTRATION_INFO, *LPREGISTRATION_INFO;

// Current connection data structure
typedef struct _CURRENT_CONNECTION_DATA
{
  SOCKET    hSocket;
  WORD      wStatus;
  BOOL		bPendingRead;
  int		iBufferLen;
} CURRENT_CONNECTION_DATA, *LPCURRENT_CONNECTION_DATA;

// Current connection structure
typedef struct _CURRENT_CONNECTION
{
  CURRENT_CONNECTION_DATA ccdIn;
  CURRENT_CONNECTION_DATA ccdOut;
} CURRENT_CONNECTION, *LPCURRENT_CONNECTION;

// Print to the console and return bytes written
DWORD printf(LPCTSTR args, ...)
{
  LPTSTR output = (LPTSTR) GlobalAlloc(GMEM_FIXED, 65536); // hopefully that's enough ;-)
 // DWORD written;
  va_list params;
  va_start(params, args);
  wvsprintf(output, args, params);
  MessageBox(HWND_DESKTOP, output, "PortProxyW95", MB_OK | MB_ICONASTERISK);
  va_end(params);
  GlobalFree(GlobalHandle(output));
  return 0;//written;
}

// Print to the socket and return bytes written
DWORD sockprintf(SOCKET hSocket, LPCTSTR args, ...)
{
  LPTSTR output = (LPTSTR) GlobalAlloc(GMEM_FIXED, 65536); // hopefully that's enough ;-)
//  DWORD written;
  va_list params;
  va_start(params, args);
  wvsprintf(output, args, params);
  send(hSocket, output, lstrlen(output), 0);
  va_end(params);
  GlobalFree(GlobalHandle(output));
  return 0;//written;
}

/* Read a line from the input stream into lpszLine, changing lpszLine's allocation as needed */
BOOL ReadTextLine(SOCKET hSocket, WSAEVENT hEvent, LPTSTR *lpszLine, LPDWORD lpdwLineLen, LPDWORD lpdwLineAlloc)
{
  // Enable blocking mode
  DWORD bBlock = false;
  WSAEventSelect(hSocket, hEvent, 0);
  ioctlsocket(hSocket, FIONBIO, (u_long FAR*) &bBlock);

  // Loop until a newline character is found
  BOOL bExit = FALSE;
  char cNext;
  *lpdwLineLen = 0;
  do
  {
    // Determine if the string allocation needs an extra character
    if (*lpdwLineLen + 2 > *lpdwLineAlloc) {
      *lpdwLineAlloc = *lpdwLineLen + 32;
      *lpszLine = (LPTSTR) GlobalReAlloc(GlobalHandle(*lpszLine), *lpdwLineAlloc, GMEM_ZEROINIT | GMEM_MOVEABLE);
    }

    /* Read in the next character */
    if (recv(hSocket, (char *) &cNext, 1, 0) != 1)
      bExit = TRUE;
    else
    {
      // If this character is a not a newline or a carriage return, then add it
      if ((cNext != 10) && (cNext != 13))
        (*lpszLine)[(*lpdwLineLen)++] = cNext;
    }
  } while (!bExit && (cNext != 10));

  // Terminate the string
  (*lpszLine)[*lpdwLineLen] = 0;

  // Disable blocking mode
  WSAEventSelect(hSocket, hEvent, FD_READ | FD_CLOSE);

  // Return whether or not more data probably exists in the file
  return !bExit;
}

// Gets a number from a string
DWORD GetNumber(LPCTSTR lpStr, DWORD *dwPos, DWORD dwLen)
{
  // Start at the given position; skip past any leading spaces
  while (*dwPos < dwLen) {
	if (lpStr[*dwPos] == ' ')
	  (*dwPos)++;
	else
	  break;
  }

  // Read in the number while we can
  DWORD dwNum = 0;
  while (*dwPos < dwLen) {
	if ((lpStr[*dwPos] >= '0') && (lpStr[*dwPos] <= '9')) {
	  dwNum = dwNum * 10 + (lpStr[*dwPos] - '0');
	  (*dwPos)++;
	} else
	  break;
  }

  // Return our number
  (*dwPos)++;
  return dwNum;	  
}

// Gets a string from a string
void GetString(LPCTSTR lpStr, DWORD *dwPos, DWORD dwLen, LPTSTR lpszOut, DWORD dwOutLen)
{
  // Start at the given position; skip past any leading spaces
  while (*dwPos < dwLen) {
	if (lpStr[*dwPos] == ' ')
	  *dwPos++;
	else
	  break;
  }

  // Read in the number while we can
  DWORD i = 0;
  lpszOut[0] = 0;
  while (*dwPos < dwLen) {
	if ((lpStr[*dwPos] != ' ') && (i < dwOutLen - 1)) {
	  lpszOut[i] = lpStr[*dwPos];
	  i++;
	  (*dwPos)++;
	} else
	  break;
  }

  // Terminate the string
  lpszOut[i] = 0;
  (*dwPos)++;
}

// Fast versions of FD_CLR and FD_SET
#define FD_FASTCLR(fd, set) do { \
    u_int __i; \
    for (__i = 0; __i < ((fd_set FAR *)(set))->fd_count ; __i++) { \
        if (((fd_set FAR *)(set))->fd_array[__i] == fd) { \
		    ((fd_set FAR *)(set))->fd_array[__i] = ((fd_set FAR *)(set))->fd_array[((fd_set FAR *)(set))->fd_count - 1]; \
            ((fd_set FAR *)(set))->fd_count--; \
            break; \
        } \
    } \
} while(0)
#define FD_FASTSET(fd, set) do { \
    if (((fd_set FAR *)(set))->fd_count < FD_SETSIZE) { \
        ((fd_set FAR *)(set))->fd_array[(((fd_set FAR *)(set))->fd_count)++] = (fd); \
    } \
} while(0)

// Definitions for event types
#define EVENTTYPE_READ   1
#define EVENTTYPE_WRITE  2
#define EVENTTYPE_EXCEPT 3

// Thread for a single TCP port (handles new and current connections simultaneously!)
DWORD WINAPI TCPPortThread(LPREGISTRATION_INFO lpInfo)
{
  // Initialize the variables to wait for incoming connections
  LPCURRENT_CONNECTION lpConnections = (LPCURRENT_CONNECTION) GlobalAlloc(GMEM_FIXED, 1);
  LPBYTE lpBuffer = (LPBYTE) GlobalAlloc(GMEM_FIXED, 65536);
  LPCURRENT_CONNECTION lpCurrent;
  DWORD nConnections = 0;
  fd_set fdRead, fdReadReturned;
  fd_set fdWrite, fdWriteReturned;
  fd_set fdExcept, fdExceptReturned;
  fd_set fdWriteCheck;
  struct timeval tvWriteCheck;
  LPCURRENT_CONNECTION_DATA lpIn, lpOut;  
  DWORD i;
  DWORD dwCurrentType;
  SOCKET hCurrentSocket;
  BOOL bBlock = true;
  BOOL bResume;
  int iXferFlags;
  int nBytes;
  memset(&tvWriteCheck, 0, sizeof(struct timeval));
  FD_ZERO(&fdRead);
  FD_ZERO(&fdWrite);
  FD_ZERO(&fdExcept);
  FD_FASTSET(lpInfo->hSocket, &fdRead);

  // Enter main port loop
  while (!lpInfo->bExit)
  {
    // Copy the current event structures into the memory we use for calls
    CopyMemory(&fdReadReturned, &fdRead, sizeof(fd_set));
    CopyMemory(&fdWriteReturned, &fdWrite, sizeof(fd_set));
    CopyMemory(&fdExceptReturned, &fdExcept, sizeof(fd_set));

    // Wait for the next event
    if (select(NULL, &fdReadReturned, &fdWriteReturned, &fdExceptReturned, NULL) == SOCKET_ERROR) {
      lpInfo->bExit = true;
      break;
    }

    // We will always have some sort of event; use post-test loop
    bResume = true;
    dwCurrentType = EVENTTYPE_WRITE;
    do
    {
      // Select the event type to process
      if ((dwCurrentType == EVENTTYPE_WRITE) && (!fdWriteReturned.fd_count))
        dwCurrentType = EVENTTYPE_READ;
      if ((dwCurrentType == EVENTTYPE_READ) && (!fdReadReturned.fd_count))
        dwCurrentType = EVENTTYPE_EXCEPT;
      if ((dwCurrentType == EVENTTYPE_EXCEPT) && (!fdExceptReturned.fd_count))
        break;

      // Get the socket to work with
      switch (dwCurrentType) {
      case EVENTTYPE_READ  : hCurrentSocket = fdReadReturned.fd_array[fdReadReturned.fd_count - 1]; (fdReadReturned.fd_count)--; break;
      case EVENTTYPE_WRITE : hCurrentSocket = fdWriteReturned.fd_array[fdWriteReturned.fd_count - 1]; (fdWriteReturned.fd_count)--; break;
      case EVENTTYPE_EXCEPT : hCurrentSocket = fdExceptReturned.fd_array[fdExceptReturned.fd_count - 1]; (fdExceptReturned.fd_count)--; break;
      }

      // Get the socket's connection information
      lpCurrent = NULL;
      for (i = 0; i < nConnections; i++)
      {
        // Check to see which socket this equals, if any
        if (hCurrentSocket == lpConnections[i].ccdIn.hSocket) {
          lpCurrent = &(lpConnections[i]);
          lpIn = &(lpConnections[i].ccdIn);
          lpOut = &(lpConnections[i].ccdOut);
          i = nConnections;
        } else {
          if (hCurrentSocket == lpConnections[i].ccdOut.hSocket) {
            lpCurrent = &(lpConnections[i]);
            lpIn = &(lpConnections[i].ccdOut);
            lpOut = &(lpConnections[i].ccdIn);
            i = nConnections;
          }
        }
      }

      // The current socket is the listening socket, then we have a new connection
      if ((hCurrentSocket == lpInfo->hSocket) && (dwCurrentType == EVENTTYPE_READ))
      {
        // Accept the incoming connection
        struct sockaddr addr;
        int addrlen = sizeof(struct sockaddr);
        SOCKET hInSocket, hOutSocket;
        if ((hInSocket = accept(lpInfo->hSocket, &addr, &addrlen)) != INVALID_SOCKET)
        {
          // Create a new connection structure member
          nConnections++;
          lpConnections = (LPCURRENT_CONNECTION) GlobalReAlloc(GlobalHandle(lpConnections), nConnections * sizeof(CURRENT_CONNECTION), GMEM_MOVEABLE);
          lpCurrent = &(lpConnections[nConnections - 1]);

          // Create a new socket for our internal connection
          hOutSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);

          // Set the sockets for non-blocking mode
          ioctlsocket(hOutSocket, FIONBIO, (u_long FAR*) &bBlock);
          ioctlsocket(hInSocket, FIONBIO, (u_long FAR*) &bBlock);

	      // Initialize the variables on this connection
          lpCurrent->ccdIn.hSocket = hInSocket;
          lpCurrent->ccdIn.wStatus = STATUS_READY;
          lpCurrent->ccdIn.bPendingRead = false;
          lpCurrent->ccdOut.hSocket = hOutSocket;
          lpCurrent->ccdOut.wStatus = STATUS_STARTING;
          lpCurrent->ccdOut.bPendingRead = false;

		  // Get the length of the write buffer on the in-port
		  i = sizeof(int);
	      getsockopt(lpCurrent->ccdIn.hSocket, SOL_SOCKET, SO_SNDBUF, (char FAR*) &(lpCurrent->ccdIn.iBufferLen), (int *) &i);
		  if (lpCurrent->ccdIn.iBufferLen > 65536)
			lpCurrent->ccdIn.iBufferLen = 65536;

          // Create a connection to the internal host
          if (connect(hOutSocket, (struct sockaddr *) &(lpInfo->saInternal), 16) != 0)
          {
            // If the error was not WSAEWOULDBLOCK, then return an error
            if (WSAGetLastError() != WSAEWOULDBLOCK) {
              printf("Could not connect to %s:%u.  Local port is %u.  Error code %lu returned.\n", lpInfo->szInternalAddress, lpInfo->wInternalPort, lpInfo->wPort, WSAGetLastError());
              closesocket(hInSocket);
              closesocket(hOutSocket);
              nConnections--;
            } else
            {
              // Set ourselves up to get a connection notification
              FD_FASTSET(hOutSocket, &fdWrite);
              FD_FASTSET(hOutSocket, &fdExcept);
            }
          } else
          {
            // We connected instantly (cool!); set ourselves up to run
            lpCurrent->ccdOut.wStatus = STATUS_READY;
            FD_FASTSET(hInSocket, &fdRead);
            FD_FASTSET(hInSocket, &fdExcept);
            FD_FASTSET(hOutSocket, &fdRead);
            FD_FASTSET(hOutSocket, &fdExcept);

			// Get the length of the write buffer on the out-port
		    i = sizeof(int);
	        getsockopt(lpCurrent->ccdOut.hSocket, SOL_SOCKET, SO_SNDBUF, (char FAR*) &(lpCurrent->ccdOut.iBufferLen), (int *) &i);
		    if (lpCurrent->ccdOut.iBufferLen > 65536)
			  lpCurrent->ccdOut.iBufferLen = 65536;
          }
        }
        continue;
      }

      // If this is a read event, then process it
      if ((dwCurrentType == EVENTTYPE_READ) || (dwCurrentType == EVENTTYPE_EXCEPT))
      {
		// Check the writability of the socket first
        FD_ZERO(&fdWriteCheck);
		FD_FASTSET(lpOut->hSocket, &fdWriteCheck);
		if (select(NULL, NULL, &fdWriteCheck, NULL, &tvWriteCheck) != SOCKET_ERROR)
		{
		  // If the output buffer is not writable, then we'll have to do some things
		  if (!fdWriteCheck.fd_count)
		  {
            // Set ourselves to wait until we can write to the output port
            FD_FASTSET(lpOut->hSocket, &fdWrite);
            FD_FASTCLR(lpIn->hSocket, &fdRead);
			lpOut->bPendingRead = true;
			break;
		  }
		}

        // If we have OOB data, then set the MSG_OOB flag
        if (dwCurrentType == EVENTTYPE_READ)
          iXferFlags = 0;
        else
          iXferFlags = MSG_OOB;

        // Determine how many bytes to read
        ioctlsocket(lpIn->hSocket, FIONREAD, (u_long FAR*) &nBytes);
        if (nBytes > lpOut->iBufferLen)
          nBytes = lpOut->iBufferLen;

        // Read the data from the input socket
        nBytes = recv(lpIn->hSocket, (char *) lpBuffer, nBytes, iXferFlags);

        // Write to the output socket if we have data
        if (nBytes > 0)
		{
		  // If we are closing the output port, then don't write
		  if (lpOut->wStatus == STATUS_CLOSING)
			break;
          if (send(lpOut->hSocket, (char *) lpBuffer, nBytes, iXferFlags) == SOCKET_ERROR)
          {
			// Abortive shutdown on the socket; do the same to the peer
			LINGER linger;
			linger.l_onoff = true;
			linger.l_linger = 0;
			setsockopt(lpOut->hSocket, SOL_SOCKET, SO_LINGER, (const char FAR *) &linger, sizeof(LINGER));
			closesocket(lpOut->hSocket);
			
			// Close the connection that aborted us
			closesocket(lpIn->hSocket);
			
			// Remove the sockets from all three arrays, if in them
			FD_FASTCLR(lpIn->hSocket, &fdRead);
			FD_FASTCLR(lpIn->hSocket, &fdWrite);
			FD_FASTCLR(lpIn->hSocket, &fdExcept);
			FD_FASTCLR(lpOut->hSocket, &fdRead);
			FD_FASTCLR(lpOut->hSocket, &fdWrite);
			FD_FASTCLR(lpOut->hSocket, &fdExcept);
			
			// Copy the last connection into the current connection
			if (lpCurrent != &(lpConnections[nConnections - 1]))
			  CopyMemory(lpCurrent, &(lpConnections[nConnections - 1]), sizeof(CURRENT_CONNECTION));
			
			// Reduce the number of connections
			nConnections--;
			
			// Force a new call to select
			bResume = false;
          }
        } else
        {
          // Check the error and dispatch it
          if (nBytes == 0)
            {
              // Ensure that we are not connecting
              if (lpCurrent->ccdOut.wStatus != STATUS_STARTING)
              {
                // We are shutting down gracefully
                if ((lpOut->wStatus == STATUS_READY) && (lpIn->wStatus == STATUS_READY))
                {
                  // We are on the first stage of the shutdown; close the send port on the output
                  shutdown(lpOut->hSocket, SD_SEND);
                  lpOut->wStatus = STATUS_CLOSING;
                  FD_FASTCLR(lpIn->hSocket, &fdRead);
                  FD_FASTCLR(lpIn->hSocket, &fdExcept);

                  // Force a new call to select
                  bResume = false;
                } else
                {
                  // We are on the second stage; set ourselves up to close when we're done writing
                  lpOut->wStatus = STATUS_CLOSING;
                  FD_FASTCLR(lpIn->hSocket, &fdRead);
                  FD_FASTCLR(lpIn->hSocket, &fdExcept);
                  FD_SET(lpOut->hSocket, &fdWrite);

                  // Force a new call to select
                  bResume = false;
                }
                break;
              }
          } else
		  {

			// Process shutdown only if we are not already in a close sequence
			// Abortive shutdown on the socket; do the same to the peer
			LINGER linger;
			linger.l_onoff = true;
			linger.l_linger = 0;
			setsockopt(lpOut->hSocket, SOL_SOCKET, SO_LINGER, (const char FAR *) &linger, sizeof(LINGER));
			closesocket(lpOut->hSocket);
			
			// Close the connection that aborted us
			closesocket(lpIn->hSocket);
			
			// Remove the sockets from all three arrays, if in them
			FD_FASTCLR(lpIn->hSocket, &fdRead);
			FD_FASTCLR(lpIn->hSocket, &fdWrite);
			FD_FASTCLR(lpIn->hSocket, &fdExcept);
			FD_FASTCLR(lpOut->hSocket, &fdRead);
			FD_FASTCLR(lpOut->hSocket, &fdWrite);
			FD_FASTCLR(lpOut->hSocket, &fdExcept);
			
			// Copy the last connection into the current connection
			if (lpCurrent != &(lpConnections[nConnections - 1]))
			  CopyMemory(lpCurrent, &(lpConnections[nConnections - 1]), sizeof(CURRENT_CONNECTION));
			
			// Reduce the number of connections
			nConnections--;
			
			// Force a new call to select
			bResume = false;
		  }
        }
      }

      // If this is a write notification, then process it (it can be a connect success, write ready, or close stage 3)
      if (dwCurrentType == EVENTTYPE_WRITE)
      {
        // If both ports are closing, then we are in the third stage of closing
        if ((lpOut->wStatus == STATUS_CLOSING) && (lpIn->wStatus == STATUS_CLOSING))
        {
          // Close the sockets
          shutdown(lpIn->hSocket, SD_SEND);
          closesocket(lpIn->hSocket);
          closesocket(lpOut->hSocket);

          // Remove the sockets from all three arrays, if in them
          FD_FASTCLR(lpIn->hSocket, &fdWrite);
          FD_FASTCLR(lpOut->hSocket, &fdWrite);
          
          // Copy the last connection into the current connection
          if (lpCurrent != &(lpConnections[nConnections - 1]))
            CopyMemory(lpCurrent, &(lpConnections[nConnections - 1]), sizeof(CURRENT_CONNECTION));

          // Reduce the number of connections
          nConnections--;

          // Force a new call to select
          bResume = false;
          break;
        }

        // If the input port is starting, then we have just connected :-)!
        if (lpIn->wStatus == STATUS_STARTING)
        {
          // Set our new ready status and remove the write event
          lpIn->wStatus = STATUS_READY;
          FD_FASTCLR(lpIn->hSocket, &fdWrite);

          // Enable the rest of the events for this connection
          FD_FASTSET(lpIn->hSocket, &fdRead);
          FD_FASTSET(lpOut->hSocket, &fdRead);
          FD_FASTSET(lpOut->hSocket, &fdExcept);

		  // Get the length of the write buffer on the out-port
		  i = sizeof(int);
	      getsockopt(lpIn->hSocket, SOL_SOCKET, SO_SNDBUF, (char FAR*) &(lpIn->iBufferLen), (int *) &i);
		  if (lpIn->iBufferLen > 65536)
			lpIn->iBufferLen = 65536;

		  // Force a new call to select
          bResume = false;
          break;
        }

        // If we have a buffer, then try writing it now (note that we are writing to input in this case)
        if (lpIn->bPendingRead)
        {
		  // Remove the write event and re-enable reading from the peer
		  FD_FASTCLR(lpIn->hSocket, &fdWrite);
		  FD_FASTSET(lpOut->hSocket, &fdRead);
		  lpIn->bPendingRead = false;
        }
      }
    } while (bResume);
  }

  // Clean up any existing connections
  for (i = 0; i < nConnections; i++)
  {
    // Get the current connection
    lpCurrent = &(lpConnections[i]);

    // Close the sockets
    shutdown(lpCurrent->ccdOut.hSocket, SD_BOTH);
    shutdown(lpCurrent->ccdIn.hSocket, SD_BOTH);
    closesocket(lpCurrent->ccdOut.hSocket);
    closesocket(lpCurrent->ccdIn.hSocket);
  }

  // Free the arrays
  GlobalFree(GlobalHandle(lpConnections));
  GlobalFree(GlobalHandle(lpBuffer));

  // Return zero
  return 0;
}

// Thread for a single UDP port (handles new and current connections simultaneously!)
DWORD WINAPI UDPPortThread(LPREGISTRATION_INFO lpInfo)
{
  // Setup an output datagram socket
  SOCKET hOutSocket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
//  SOCKET hOutSocket = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
  int bOn = 1;
  if (hOutSocket == INVALID_SOCKET) {
	printf("invalid socket");
	return 0;
  }
//printf("start, udp");
//  if(setsockopt(hOutSocket, IPPROTO_IP, 3, (char *) &bOn, sizeof(int)) != 0)
//	printf("ssopt!");

  // Create the exchange buffer based on the size of the largest message
  DWORD dwBufferLen;
  int iSize = sizeof(DWORD);
  if (getsockopt(lpInfo->hSocket, SOL_SOCKET, SO_MAX_MSG_SIZE, (char FAR*) &dwBufferLen, &iSize)) {
	printf("Could not get the buffer length on UDP port %u!\n", lpInfo->wPort);
	closesocket(hOutSocket);
	return 0;
  }

  // Allocate the memory for the buffer
  LPBYTE lpBuffer = (LPBYTE) GlobalAlloc(GMEM_FIXED, dwBufferLen + sizeof(struct iphdr) + sizeof(struct udphdr));
  if (!lpBuffer) {
	printf("Could not allocate buffer space on UDP port %u!\n", lpInfo->wPort);
	closesocket(hOutSocket);
	return 0;
  }

  // Continually forward UDP messages from the server port to the destination
  struct sockaddr_in addr;
//  struct udphdr *udphdr;
  WORD dwSeq = 1;
  int addrlen;
  int nBytes;
  while (!lpInfo->bExit)
  {
	// Make a blocking read call to get the next packet
	addrlen = sizeof(struct sockaddr); /* optional? */
	nBytes = recvfrom(lpInfo->hSocket, (char FAR*) (lpBuffer + sizeof(struct iphdr) + sizeof(struct udphdr)), dwBufferLen, 0, (struct sockaddr*) &addr, &addrlen);
//	nBytes = recvfrom(lpInfo->hSocket, (char FAR*) (lpBuffer), dwBufferLen, 0, (struct sockaddr*) &addr, &addrlen);
//printf("(udp)->");

/*	if (!lpInfo->bExit && nBytes && (nBytes != SOCKET_ERROR))
	{
      // Generate the IP header
      ip_gen((char *) lpBuffer, IPPROTO_UDP, addr.sin_addr, lpInfo->saInternal.sin_addr, nBytes + sizeof(struct iphdr) + sizeof(struct udphdr), &dwSeq);

	  // Generate the UDP header
      udphdr = (struct udphdr *)(lpBuffer + sizeof(struct iphdr));
      udp_gen((char *)udphdr, addr.sin_port, lpInfo->saInternal.sin_port, nBytes + sizeof(struct udphdr));
      udphdr->check = trans_check(IPPROTO_UDP, (char *) udphdr, nBytes + sizeof(struct udphdr), addr.sin_addr, lpInfo->saInternal.sin_addr);
//      udphdr->check = in_cksum((unsigned short*) lpBuffer, nBytes + sizeof(struct iphdr) + sizeof(struct udphdr));

	  // Send the UDP/IP packet
	  if ((nBytes = sendto(hOutSocket, (char FAR*) lpBuffer, nBytes + sizeof(struct iphdr) + sizeof(struct udphdr), 0, (struct sockaddr *) &(lpInfo->saInternal), sizeof(struct sockaddr))) == SOCKET_ERROR) {
		printf("Error writing UDP packet on port %u.  Error code %lu returned.  Check the internal server.\n", lpInfo->wPort, WSAGetLastError());
		lpInfo->bExit = true;
	  } else
		printf("(%lu)", nBytes);*/
	// Send this packet to the destination if we're not exiting
	if (!lpInfo->bExit && nBytes && (nBytes != SOCKET_ERROR)) {
	  if (sendto(hOutSocket, (char FAR*) lpBuffer, nBytes, 0, (struct sockaddr *) &(lpInfo->saInternal), 16) == SOCKET_ERROR) {
		printf("Error writing UDP packet on port %u.  Check the internal server.\n", lpInfo->wPort);
		lpInfo->bExit = true;
	  }
	}// else
//	  printf("ohdear");
  }
//intf("done, udp");
  // Free the buffer, close the output socket, and leave
  GlobalFree(GlobalHandle(lpBuffer));
  closesocket(hOutSocket);
  return 0;
}

// Service handler routine
/*void WINAPI ServiceHandlerProc(DWORD dwControl)
{
  // We only process the stop message...
  if (dwControl == SERVICE_CONTROL_STOP)
  {
    // Set the bExit flag to true; it will automatically close within the next second :-)
    bExit = true;

	// Wait for the main loop to finish, then return
	EnterCriticalSection(&csMainLoop);
	LeaveCriticalSection(&csMainLoop);
    DeleteCriticalSection(&csMainLoop);
  }
} 

// Handler for console window signals
BOOL WINAPI SignalHandler(DWORD dwCtrlType)
{
  // If this function is called, the user pressed Ctrl+C, pressed Ctrl+Break, clicked
  // on the close box, selected Close from the system menu, told Windows to logout,
  // or told Windows to shut down the system.  In all cases, we set a signal that
  // ensures that the program is shutdown gracefully while in debug mode

  // Call ServiceHandlerProc with SERVICE_CONTROL_STOP
  ServiceHandlerProc(SERVICE_CONTROL_STOP);
  
  // Return true now that we're shutting down
  return true;
}*/

// Function that determines if a command-line flag was set
BOOL IsFlagSet(LPTSTR *lpszCharList, LPTSTR lpszFlagName)
{
  /* Is the requested flag set in the command line? */
  DWORD i = 1;
  BOOL bFound = FALSE, bInParamList = TRUE;
  while (lpszCharList[i] && !bFound && bInParamList)
  {
    if ((lpszCharList[i][0] != '-') && (lpszCharList[i][0] != '/'))
      bInParamList = FALSE;
    else
    {
      if (lstrcmpi(lpszCharList[i], lpszFlagName) == 0)
        bFound = TRUE;
      else 
        i++;
    }
  }
  return bFound;
}

// Service main entry point, as called from the service manager
DWORD WINAPI ServiceMain(LPTSTR *lpszArgv)
{
  DWORD dwArgc = 1;
  // Create a critical section object and enter it
/*  InitializeCriticalSection(&csMainLoop);
  EnterCriticalSection(&csMainLoop);

  // Register the service control handler function if we're running as a service
  SERVICE_STATUS_HANDLE sshService = NULL;
  SERVICE_STATUS stat;
  ZeroMemory(&stat, sizeof(SERVICE_STATUS));
  bExit = false;
  if (!bDebugMode)
  {
    if (!(sshService = RegisterServiceCtrlHandler(SERVICE_NAME, (LPHANDLER_FUNCTION) ServiceHandlerProc))) {
	  printf("Call to RegisterServiceCtrlHandler for %s failed!  Error code %lu returned.\n", SERVICE_NAME, GetLastError());
	  LeaveCriticalSection(&csMainLoop);
      DeleteCriticalSection(&csMainLoop);
	  return;
	}
  } else
  {
    // Set the console to have us handle signals instead of the system :-)
    SetConsoleCtrlHandler((PHANDLER_ROUTINE) SignalHandler, true);
  }

  // Set the service status to loading
  if (!bDebugMode) {
	stat.dwServiceType = SERVICE_WIN32_OWN_PROCESS;
	stat.dwCurrentState = SERVICE_START_PENDING;
	stat.dwControlsAccepted = SERVICE_ACCEPT_STOP;
	stat.dwWin32ExitCode = NO_ERROR;
	stat.dwServiceSpecificExitCode = 0;
	stat.dwCheckPoint = 0;
	stat.dwWaitHint = 1000; // shouldn't take more than one second per BIND
	SetServiceStatus(sshService, &stat);
  }*/

  // Open the registry key for the service and read the registration info
  HKEY hKey;
  DWORD nRegistrations = 0;
  LPREGISTRATION_INFO lpRegistrations = NULL, lpCurrent;
  DWORD i, dwThreadID;
  if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, "SYSTEM\\CurrentControlSet\\Services\\PortProxy\\Parameters", 0, KEY_QUERY_VALUE, &hKey) == ERROR_SUCCESS)
  {
	// Get the number of registrations
	DWORD dwLen = sizeof(DWORD);
	DWORD dwType = REG_DWORD;
	if (RegQueryValueEx(hKey, "NumRegistrations", NULL, &dwType, (LPBYTE) &nRegistrations, &dwLen) != ERROR_SUCCESS)
	  nRegistrations = 0;

	// Read each registration
	lpCurrent = NULL;
	if (nRegistrations) {
	  char szValue[30];
	  DWORD dwValue;
	  for (i = 0; i < nRegistrations; i++)
	  {
		// Add a new linked list member
		if (lpCurrent) {
		  lpCurrent->next = GlobalAlloc(GPTR, sizeof(REGISTRATION_INFO));
		  lpCurrent = (LPREGISTRATION_INFO) lpCurrent->next;
		} else {
		  lpRegistrations = (LPREGISTRATION_INFO) GlobalAlloc(GPTR, sizeof(REGISTRATION_INFO));
		  lpCurrent = lpRegistrations;
		}

		// Get the type of port
		dwLen = sizeof(DWORD);
		dwType = REG_DWORD;
		wsprintf(szValue, "Type%lu", i);
	    if (RegQueryValueEx(hKey, szValue, NULL, &dwType, (LPBYTE) &dwValue, &dwLen) != ERROR_SUCCESS)
		  lpCurrent->dwType = IPPROTO_TCP;
		else
		  lpCurrent->dwType = (dwValue == IPPROTO_UDP)? IPPROTO_UDP: IPPROTO_TCP;

  	    // Get the local port
	    dwLen = sizeof(DWORD);
		dwType = REG_DWORD;
		wsprintf(szValue, "Port%lu", i);
	    if (RegQueryValueEx(hKey, szValue, NULL, &dwType, (LPBYTE) &dwValue, &dwLen) != ERROR_SUCCESS)
		  lpCurrent->wPort = 0;
		else
		  lpCurrent->wPort = (WORD) dwValue;

		// Get the internal port
	    dwLen = sizeof(DWORD);
		dwType = REG_DWORD;
		wsprintf(szValue, "InternalPort%lu", i);
	    if (RegQueryValueEx(hKey, szValue, NULL, &dwType, (LPBYTE) &dwValue, &dwLen) != ERROR_SUCCESS)
		  lpCurrent->wInternalPort = 0;
		else
		  lpCurrent->wInternalPort = (WORD) dwValue;

		// Get the internal address
	    dwLen = sizeof(char) * 16;
		dwType = REG_SZ;
		wsprintf(szValue, "InternalAddress%lu", i);
	    if (RegQueryValueEx(hKey, szValue, NULL, &dwType, (LPBYTE) lpCurrent->szInternalAddress, &dwLen) != ERROR_SUCCESS)
		  lpCurrent->szInternalAddress[0] = 0;

		// Set the dwRegId to the current number
		lpCurrent->dwRegId = i;

		// Ensure the "next" variable is NULL (should have been set by GlobalAlloc, but this makes sure ;-)
		lpCurrent->next = NULL;
	  }
	}

	// Close the key
	RegCloseKey(hKey);
  }

  // Process each registration
  struct sockaddr_in pAddr;
  ZeroMemory(&pAddr, sizeof(struct sockaddr_in));
  pAddr.sin_family = AF_INET;
  pAddr.sin_addr.s_addr = INADDR_ANY;
  lpCurrent = lpRegistrations;
  while (lpCurrent)
  {
	// Update the service manager on our status
/*    if (!bDebugMode) {
      stat.dwCheckPoint++;
      SetServiceStatus(sshService, &stat);
	}*/

    // Create a listening socket for this port
    lpCurrent->hSocket = socket(AF_INET, (lpCurrent->dwType == IPPROTO_UDP)? SOCK_DGRAM: SOCK_STREAM, lpCurrent->dwType);

	// Bind the socket to its port
	pAddr.sin_port = htons(lpCurrent->wPort);
	if (bind(lpCurrent->hSocket, (struct sockaddr *) &pAddr, sizeof(pAddr))) {
  	  printf("Could not bind to port %lu.\n", lpCurrent->wPort);
	  shutdown(lpCurrent->hSocket, SD_BOTH);
	  closesocket(lpCurrent->hSocket);
	  lpCurrent->hSocket = NULL;
	  lpCurrent->hThread = NULL;
	} else
	{
	  // Initiate listening if we're stream-oriented
	  if (lpCurrent->dwType != IPPROTO_UDP)
	    listen(lpCurrent->hSocket, SOMAXCONN);
	  
	  // Generate a socket address for the internal computer
	  memset(&(lpCurrent->saInternal), 0, sizeof(struct sockaddr_in));
	  lpCurrent->saInternal.sin_family = AF_INET;
	  lpCurrent->saInternal.sin_addr.s_addr = inet_addr(lpCurrent->szInternalAddress);
	  lpCurrent->saInternal.sin_port = htons(lpCurrent->wInternalPort);
	  
	  // Create the thread for this port and run it
	  lpCurrent->bExit = false;
	  lpCurrent->hThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE) ((lpCurrent->dwType == IPPROTO_UDP)? UDPPortThread: TCPPortThread), lpCurrent, 0, &dwThreadID);
	  if (!lpCurrent->hThread) {
		printf("Could not create thread for port %lu!  Error code is %lu returned.\n", lpCurrent->wPort, GetLastError());
		if (lpCurrent->dwType != IPPROTO_UDP)
		  shutdown(lpCurrent->hSocket, SD_BOTH);
		closesocket(lpCurrent->hSocket);
		lpCurrent->hSocket = NULL;
	  }
	}

	// Next registration
	lpCurrent = (LPREGISTRATION_INFO) lpCurrent->next;
  }

  // Update the service manager on our status
/*  if (!bDebugMode) {
    stat.dwCheckPoint++;
    SetServiceStatus(sshService, &stat);
  }*/

  // Create a local server for dynamic configuration modifications
  SOCKET hConfigSocket;
  WSAEVENT hConfigEvent;
  hConfigSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  pAddr.sin_port = htons(256);
  bind(hConfigSocket, (struct sockaddr *) &pAddr, sizeof(pAddr));
  listen(hConfigSocket, SOMAXCONN);
  hConfigEvent = WSACreateEvent();
  WSAEventSelect(hConfigSocket, hConfigEvent, FD_ACCEPT);

  // Return the pAddr structure to its previous state
  pAddr.sin_family = AF_INET;
  pAddr.sin_addr.s_addr = INADDR_ANY;

  // Update the service manager on our running status
/*  if (!bDebugMode) {
    stat.dwCheckPoint = 0;
    stat.dwCurrentState = SERVICE_RUNNING;
    SetServiceStatus(sshService, &stat);
  }*/

  // Enter program loop
  WSAEVENT *lpEvents = (WSAEVENT *) GlobalAlloc(GMEM_FIXED, sizeof(WSAEVENT));
  SOCKET *lpSockets = (SOCKET *) GlobalAlloc(GMEM_FIXED, 1);
  LPTSTR lpszLine = (LPTSTR) GlobalAlloc(GMEM_FIXED, 32);
  DWORD dwLineLen = 0, dwLineAlloc = 32;
  DWORD nEvents = 1, nSockets = 0;
  DWORD curEvent;
  WSANETWORKEVENTS netEvent;
  lpEvents[0] = hConfigEvent;
  while (!bExit)
  {
	WSASetLastError(0);
	curEvent = WSAWaitForMultipleEvents(nEvents, lpEvents, false, 1000, false);
	if ((curEvent != WSA_WAIT_TIMEOUT) && !bExit)
	{
	  // If we are in an error condition, then exit
	  if (WSAGetLastError()) {
		printf("WSAWaitForMultipleEvents returned %lu in main program loop and set the last error to %lu.\n", curEvent, WSAGetLastError());
		bExit = true;
		break;
	  }

	  // If the event is the first one, then we have a new connection :-)!
	  if (curEvent == 0)
	  {
		struct sockaddr_in addr;
		int addrlen = sizeof(struct sockaddr);
		SOCKET hNewSocket;

		// Get the network event
	    WSAEnumNetworkEvents(hConfigSocket, hConfigEvent, &netEvent);

		// Accept the connection
		if ((hNewSocket = accept(hConfigSocket, (struct sockaddr *) &addr, &addrlen)) != INVALID_SOCKET)
		{
		  // Check the IP address of this connection (it MUST be 127.0.0.1; NO EXCEPTIONS!)
		  if (lstrcmp(inet_ntoa(addr.sin_addr), "127.0.0.1") == 0)
		  {
			// Add this connection to the socket array
			nSockets++;
			lpSockets = (SOCKET *) GlobalReAlloc(GlobalHandle(lpSockets), nSockets * sizeof(SOCKET), GMEM_MOVEABLE);
			nEvents++;
			lpEvents = (WSAEVENT *) GlobalReAlloc(GlobalHandle(lpEvents), nEvents * sizeof(WSAEVENT), GMEM_MOVEABLE);
			lpSockets[nSockets - 1] = hNewSocket;
			lpEvents[nEvents - 1] = WSACreateEvent();
			WSAEventSelect(lpSockets[nSockets - 1], lpEvents[nEvents - 1], FD_READ | FD_CLOSE);
		  } else
		  {
			// Produce a security error message and close the socket
			printf("Computer at %s attempted to access administrative port.  Connection was dropped.\n", inet_ntoa(addr.sin_addr));
			sockprintf(hNewSocket, "-Access denied\r\n");
			shutdown(hNewSocket, SD_BOTH);
			closesocket(hNewSocket);
		  }
		} else
		  printf("Could not accept socket!  %lu\n", WSAGetLastError());
	  } else
	  {
		// Get the network event
	    WSAEnumNetworkEvents(lpSockets[curEvent - 1], lpEvents[curEvent], &netEvent);

		// Read the command from the client
		char cmd;
		if (!recv(lpSockets[curEvent - 1], &cmd, 1, 0))
		{
		  // Remove this connection from the arrays after shutting it down
		  WSACloseEvent(lpEvents[curEvent]);
		  shutdown(lpSockets[curEvent - 1], SD_BOTH);
		  closesocket(lpSockets[curEvent - 1]);
		  
		  // Remove the event and socket from the arrays
		  if (curEvent < nEvents - 1)
			lpEvents[curEvent] = lpEvents[nEvents - 1];
		  nEvents--;
		  if (curEvent < nSockets)
			lpSockets[curEvent - 1] = lpSockets[nSockets - 1];
		  nSockets--;
		} else
		{
		  // Switch according to the command requested
		  switch(cmd) {

		  case '+' :
			{
			  // Add a registration in this form: port intaddr intport\r\n
			  if (!ReadTextLine(lpSockets[curEvent - 1], lpEvents[curEvent], &lpszLine, &dwLineLen, &dwLineAlloc))
			    break;
			  LPREGISTRATION_INFO lpNew = (LPREGISTRATION_INFO) GlobalAlloc(GPTR, sizeof(REGISTRATION_INFO));
			  DWORD dwPos = 0;
			  BOOL bSuccess = false;
              lpNew->wPort = (WORD) GetNumber(lpszLine, &dwPos, dwLineLen);
              GetString(lpszLine, &dwPos, dwLineLen, lpNew->szInternalAddress, 16);
              lpNew->wInternalPort = (WORD) GetNumber(lpszLine, &dwPos, dwLineLen);
              lpNew->dwType = GetNumber(lpszLine, &dwPos, dwLineLen);

			  // Add to the registry
			  if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, "SYSTEM\\CurrentControlSet\\Services\\PortProxy\\Parameters", 0, KEY_SET_VALUE, &hKey) == ERROR_SUCCESS)
			  {
				// Set the new number of registrations
				nRegistrations++;
				if (RegSetValueEx(hKey, "NumRegistrations", NULL, REG_DWORD, (LPBYTE) &nRegistrations, sizeof(DWORD)) == ERROR_SUCCESS)
				{
				  // Set the local port number
				  char szValue[30];
				  DWORD dwValue;
				  bSuccess = true;
				  wsprintf(szValue, "Port%lu", nRegistrations - 1);
				  dwValue = lpNew->wPort;
				  bSuccess = bSuccess && (RegSetValueEx(hKey, szValue, NULL, REG_DWORD, (LPBYTE) &dwValue, sizeof(DWORD)) == ERROR_SUCCESS);
				  
			  	  // Set the port type
				  wsprintf(szValue, "Type%lu", nRegistrations - 1);
				  dwValue = lpNew->dwType;
				  RegSetValueEx(hKey, szValue, NULL, REG_DWORD, (LPBYTE) &dwValue, sizeof(DWORD));

				  // Set the internal port
				  wsprintf(szValue, "InternalPort%lu", nRegistrations - 1);
				  dwValue = lpNew->wInternalPort;
				  bSuccess = bSuccess && (RegSetValueEx(hKey, szValue, NULL, REG_DWORD, (LPBYTE) &dwValue, sizeof(DWORD)) == ERROR_SUCCESS);
				  
				  // Set the internal address
				  wsprintf(szValue, "InternalAddress%lu", nRegistrations - 1);
				  bSuccess = bSuccess && (RegSetValueEx(hKey, szValue, NULL, REG_SZ, (LPBYTE) lpNew->szInternalAddress, lstrlen(lpNew->szInternalAddress)) == ERROR_SUCCESS);

				  // Set the dwRegId value for this registration
				  lpNew->dwRegId = nRegistrations - 1;

				  // If there was an error, then decrement nRegistrations and reset it
				  if (!bSuccess) {
					nRegistrations--;
					RegSetValueEx(hKey, "NumRegistrations", NULL, REG_DWORD, (LPBYTE) &nRegistrations, sizeof(DWORD));
				  }
				} else
				  nRegistrations--;
			  } else
				hKey = NULL;

			  // Continue only if the registry keys were written correctly
			  if (bSuccess)
			  {
				// Create a listening socket for this port
				lpNew->hSocket = socket(AF_INET, (lpNew->dwType == IPPROTO_UDP)? SOCK_DGRAM: SOCK_STREAM, lpNew->dwType);
				pAddr.sin_port = htons(lpNew->wPort);
                pAddr.sin_addr.s_addr = INADDR_ANY;
				if (bind(lpNew->hSocket, (struct sockaddr *) &pAddr, sizeof(pAddr))) {
				  printf("Could not bind to port %lu.\n", lpNew->wPort);
				  sockprintf(lpSockets[curEvent - 1], "-Could not bind to port %lu.  It is probably in use.\r\n", lpNew->wPort);
				  if (lpNew->dwType != IPPROTO_UDP)
			        shutdown(lpNew->hSocket, SD_BOTH);
				  closesocket(lpNew->hSocket);
				  nRegistrations--;
				  RegSetValueEx(hKey, "NumRegistrations", NULL, REG_DWORD, (LPBYTE) &nRegistrations, sizeof(DWORD));
				  GlobalFree(GlobalHandle(lpNew));
				} else
				{
				  // Begin listening
				  if (lpNew->dwType != IPPROTO_UDP)
				    listen(lpNew->hSocket, SOMAXCONN);
				  
				  // Generate a socket address for the internal computer
				  memset(&(lpNew->saInternal), 0, sizeof(struct sockaddr_in));
				  lpNew->saInternal.sin_family = AF_INET;
				  lpNew->saInternal.sin_addr.s_addr = inet_addr(lpNew->szInternalAddress);
				  lpNew->saInternal.sin_port = htons(lpNew->wInternalPort);
				  
				  // Create the thread for this port and run it
				  lpNew->bExit = false;
				  lpNew->hThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE) ((lpNew->dwType == IPPROTO_UDP)? UDPPortThread: TCPPortThread), lpNew, 0, &dwThreadID);
				  if (!lpNew->hThread) {
					printf("Could not create thread for port %lu!  Error code is %lu returned.\n", lpCurrent->wPort, GetLastError());
					sockprintf(lpSockets[curEvent - 1], "-Could not create thread for port %lu!  Error code is %lu returned.\r\n", lpCurrent->wPort, GetLastError());
					if (lpNew->dwType != IPPROTO_UDP)
					  shutdown(lpNew->hSocket, SD_BOTH);
					closesocket(lpNew->hSocket);
					nRegistrations--;
					RegSetValueEx(hKey, "NumRegistrations", NULL, REG_DWORD, (LPBYTE) &nRegistrations, sizeof(DWORD));
					GlobalFree(GlobalHandle(lpNew));
				  } else
				  {
					// Success :-)!  Add this to the lpRegistrations linked list (the easy way ;-)
					lpNew->next = lpRegistrations;
					lpRegistrations = lpNew;
					sockprintf(lpSockets[curEvent - 1], "+Registration active\r\n");
				  }
				}
			  } else
			  {
				// Clean up the allocation and return an error about the registry
				GlobalFree(GlobalHandle(lpNew));
				printf("Could not create registry key!  Check the permissions on HKEY_LOCAL_MACHINE\\SYSTEM\\CurrentControlSet\\Services\\PortProxy\\Parameters and ensure this service has KEY_QUERY_VALUE and KEY_SET_VALUE rights on the key.\n");
				sockprintf(lpSockets[curEvent - 1], "-Could not create registry key!  Ask administrator to check permissions.\r\n");
			  }

			  // Close the key if it's open
			  if (hKey)
			    RegCloseKey(hKey);
			  break;
			}

		  case '-' :
			{
			  // Remove a registration in this form: port\r\n
			  if (!ReadTextLine(lpSockets[curEvent - 1], lpEvents[curEvent], &lpszLine, &dwLineLen, &dwLineAlloc))
			    break;

			  // Try to find the registration in the linked list
			  LPREGISTRATION_INFO lpHighest = lpRegistrations;
			  BOOL bFound = false;
			  DWORD dwPos = 0;
              WORD wPort = (WORD) GetNumber(lpszLine, &dwPos, dwLineLen);
              DWORD dwType = GetNumber(lpszLine, &dwPos, dwLineLen);
			  lpCurrent = lpRegistrations;
			  while (lpCurrent && !bFound) {
				if (lpCurrent->dwRegId > lpHighest->dwRegId)
				  lpHighest = lpCurrent;
				if ((lpCurrent->wPort == wPort) && (lpCurrent->dwType == dwType))
				  bFound = true;
				else
				  lpCurrent = (LPREGISTRATION_INFO) lpCurrent->next;
			  }

			  // If we found it, then remove it
			  if (bFound)
			  {
				// If this one has an active thread, then process it
				if (lpCurrent->hThread)
				{
				  // Set the exit flag on the registration
				  SOCKET hCloseSocket;
				  lpCurrent->bExit = true;
				  
				  // Make a connection to the port
				  pAddr.sin_port = htons(lpCurrent->wPort);
                  pAddr.sin_addr.s_addr = inet_addr("127.0.0.1");
				  if (lpCurrent->dwType != IPPROTO_UDP) {
				    hCloseSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
				    connect(hCloseSocket, (struct sockaddr *) &pAddr, 16);
				  } else {
				    hCloseSocket = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
					sendto(hCloseSocket, (char FAR*) &hCloseSocket, sizeof(SOCKET), 0, (struct sockaddr *) &pAddr, 16);
				  }
				  
				  // Close the connection and wait for the thread to exit (up to 1 second)
				  closesocket(hCloseSocket);
				  WaitForSingleObject(lpCurrent->hThread, 1000);
				  
				  // Close the event, socket, and thread
				  if (lpCurrent->dwType != IPPROTO_UDP)
			        shutdown(lpCurrent->hSocket, SD_BOTH);
				  closesocket(lpCurrent->hSocket);
				  CloseHandle(lpCurrent->hThread);
				}

				// Open the registry so we can change it :-)
				if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, "SYSTEM\\CurrentControlSet\\Services\\PortProxy\\Parameters", 0, KEY_SET_VALUE, &hKey) == ERROR_SUCCESS)
				{
				  // Set the new number of registrations
				  char szValue[30];
				  nRegistrations--;
				  RegSetValueEx(hKey, "NumRegistrations", NULL, REG_DWORD, (LPBYTE) &nRegistrations, sizeof(DWORD));
				  
				  // Make our dwRegId the last one so we can pop off easily :-), unless if we are it
				  if (lpHighest->dwRegId != lpCurrent->dwRegId) {
					lpHighest->dwRegId = lpCurrent->dwRegId;
					
					// Set the local port number
				    DWORD dwValue;
					wsprintf(szValue, "Port%lu", lpHighest->dwRegId);
					dwValue = lpHighest->wPort;
					RegSetValueEx(hKey, szValue, NULL, REG_DWORD, (LPBYTE) &dwValue, sizeof(DWORD));
					
					// Set the port type
					wsprintf(szValue, "Type%lu", lpHighest->dwRegId);
					dwValue = lpHighest->dwType;
					RegSetValueEx(hKey, szValue, NULL, REG_DWORD, (LPBYTE) &dwValue, sizeof(DWORD));

					// Set the internal port
					wsprintf(szValue, "InternalPort%lu", lpHighest->dwRegId);
					dwValue = lpHighest->wInternalPort;
					RegSetValueEx(hKey, szValue, NULL, REG_DWORD, (LPBYTE) &dwValue, sizeof(DWORD));
					
					// Set the internal address
					wsprintf(szValue, "InternalAddress%lu", lpHighest->dwRegId);
					RegSetValueEx(hKey, szValue, NULL, REG_SZ, (LPBYTE) lpHighest->szInternalAddress, lstrlen(lpHighest->szInternalAddress));
				  }

				  // Delete the local port number of the last entry
				  wsprintf(szValue, "Port%lu", lpHighest->dwRegId);
				  RegDeleteValue(hKey, szValue);
				  
				  // Delete the type
				  wsprintf(szValue, "Type%lu", lpHighest->dwRegId);
				  RegDeleteValue(hKey, szValue);

				  // Delete the internal port
				  wsprintf(szValue, "InternalPort%lu", lpHighest->dwRegId);
				  RegDeleteValue(hKey, szValue);
				  
				  // Delete the internal address
				  wsprintf(szValue, "InternalAddress%lu", lpHighest->dwRegId);
				  RegDeleteValue(hKey, szValue);
				  
				  // Close the key
				  RegCloseKey(hKey);
				}

				// Remove this registration from the linked list
				lpHighest = lpRegistrations;
				if (lpHighest == lpCurrent)
				{
				  // We're at the top; this is easy ;-)
				  lpRegistrations = (LPREGISTRATION_INFO) lpCurrent->next;
				  GlobalFree(lpCurrent);
				} else
				{
				  // Search for ourselves and remove when found
				  bFound = false;
				  while (lpHighest && !bFound) {
				    if ((LPREGISTRATION_INFO) lpHighest->next == lpCurrent)
					{
					  // Here we are :-)!  Now let's get outta here!
					  lpHighest->next = lpCurrent->next;
					  GlobalFree(lpCurrent);
					  bFound = true;
					}
				  }

				  // If we weren't found, then alert an error to the computer but not the remote client
				  if (!bFound)
					printf("Could not remove a deleted registration from the lpRegistrations linked list!  This should NEVER happen, so please contact slawrance@technologist.com about this.  Thanks :-).\n");
				}

				// Return our deletion status
				sockprintf(lpSockets[curEvent - 1], "+Registration deleted\r\n");
			  } else
			  {
				// Registration was already deleted
				sockprintf(lpSockets[curEvent - 1], "-Registration not found or already deleted\r\n");
			  }
			  break;
			}

		  case 'x' :
			{
			  // Close this connection; first, send a response to the client, then close up
			  if (!ReadTextLine(lpSockets[curEvent - 1], lpEvents[curEvent], &lpszLine, &dwLineLen, &dwLineAlloc))
			    break;
			  sockprintf(lpSockets[curEvent - 1], "+Session closed\r\n");

		      // Remove this connection from the arrays after shutting it down
			  WSACloseEvent(lpEvents[curEvent]);
			  shutdown(lpSockets[curEvent - 1], SD_BOTH);
			  closesocket(lpSockets[curEvent - 1]);
			  
			  // Remove the event and socket from the arrays
			  if (curEvent < nEvents - 1)
				lpEvents[curEvent] = lpEvents[nEvents - 1];
			  nEvents--;
			  if (curEvent < nSockets)
				lpSockets[curEvent - 1] = lpSockets[nSockets - 1];
			  nSockets--;
			  break;
			}

		  case 's' :
			{
			  // Get the status of a registration in this form: port\r\n
			  if (!ReadTextLine(lpSockets[curEvent - 1], lpEvents[curEvent], &lpszLine, &dwLineLen, &dwLineAlloc))
			    break;

			  // Try to find the registration in the linked list
			  LPREGISTRATION_INFO lpHighest = lpRegistrations;
			  BOOL bFound = false;
			  DWORD dwPos = 0;
              WORD wPort = (WORD) GetNumber(lpszLine, &dwPos, dwLineLen);
			  lpCurrent = lpRegistrations;
			  while (lpCurrent && !bFound) {
				if (lpCurrent->dwRegId > lpHighest->dwRegId)
				  lpHighest = lpCurrent;
				if (lpCurrent->wPort == wPort)
				  bFound = true;
				else
				  lpCurrent = (LPREGISTRATION_INFO) lpCurrent->next;
			  }

			  // If we found it, then check for a running thread
			  if (bFound)
			  {
				if (lpCurrent->hThread)
				  sockprintf(lpSockets[curEvent - 1], "+Registration active\r\n");
				else
				  sockprintf(lpSockets[curEvent - 1], "-Registration exists but is not active\r\n");
			  } else
				sockprintf(lpSockets[curEvent - 1], "-Registration not found\r\n");
			  break;
			}

		  case 't' :
			{
			  // Test a server port in the form: port\r\n
			  if (!ReadTextLine(lpSockets[curEvent - 1], lpEvents[curEvent], &lpszLine, &dwLineLen, &dwLineAlloc))
			    break;

			  // Get the port that we'll try to BIND to
			  DWORD dwPos = 0;
              WORD wPort = (WORD) GetNumber(lpszLine, &dwPos, dwLineLen);

			  // Create a new socket and try to BIND to it
			  SOCKET hSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
			  pAddr.sin_port = htons(wPort);
              pAddr.sin_addr.s_addr = INADDR_ANY;
			  if (bind(hSocket, (sockaddr *) &pAddr, sizeof(pAddr)))
				sockprintf(lpSockets[curEvent - 1], "-Could not BIND; port might be in use.  Error code %lu returned.\r\n", WSAGetLastError());
			  else
				sockprintf(lpSockets[curEvent - 1], "+Port available at the moment.\r\n");
			  
			  // Delete the socket
			  shutdown(hSocket, SD_BOTH);
			  closesocket(hSocket);
			  break;
			}

		  case 'n':
			{
			  // No-op; return a positive response
			  if (!ReadTextLine(lpSockets[curEvent - 1], lpEvents[curEvent], &lpszLine, &dwLineLen, &dwLineAlloc))
			    break;
			  sockprintf(lpSockets[curEvent - 1], "+No-op to you too\r\n");
			  break;
			}
		  }
		}
	  }
	}
  }

  // Leave the critical section (it will be deleted by our service control manager event handler)
//  LeaveCriticalSection(&csMainLoop);

  // Update the service manager on our closing status
/*  if (!bDebugMode) {
    stat.dwCheckPoint = 0;
    stat.dwCurrentState = SERVICE_STOP_PENDING;
    stat.dwWaitHint = 4000; // shouldn't take more than four seconds per registration
    SetServiceStatus(sshService, &stat);
  }*/

  // Close all administrative connections, if any
  while (nSockets) {
	WSACloseEvent(lpEvents[nSockets]);
	shutdown(lpSockets[nSockets - 1], SD_BOTH);
    closesocket(lpSockets[nSockets - 1]);
	nSockets--;

    // Update the service manager on our closing status
/*    if (!bDebugMode) {
      stat.dwCheckPoint++;
      SetServiceStatus(sshService, &stat);
	}*/
  }

  // Close the administrative event and socket
  WSACloseEvent(hConfigEvent);
  shutdown(hConfigSocket, SD_BOTH);
  closesocket(hConfigSocket);

  // Free the memory associated with the events, sockets, and the line
  GlobalFree(GlobalHandle(lpEvents));
  GlobalFree(GlobalHandle(lpSockets));
  GlobalFree(GlobalHandle(lpszLine));

  // Close all the active port threads politely
  lpCurrent = lpRegistrations;
  pAddr.sin_addr.s_addr = inet_addr("127.0.0.1");
  while (lpCurrent)
  {
    // Update the service manager on our closing status
/*    if (!bDebugMode) {
      stat.dwCheckPoint++;
      SetServiceStatus(sshService, &stat);
	}*/

    // If this one has an active thread, then process it
	if (lpCurrent->hThread)
	{
	  // Set the exit flag on the registration
	  SOCKET hCloseSocket;
	  lpCurrent->bExit = true;

	  // Make a connection to the port
      pAddr.sin_port = htons(lpCurrent->wPort);
      hCloseSocket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
      connect(hCloseSocket, (struct sockaddr *) &pAddr, 16);

	  // Close the connection and wait for the thread to exit (up to 1 second)
	  shutdown(hCloseSocket, SD_BOTH);
	  closesocket(hCloseSocket);
	  WaitForSingleObject(lpCurrent->hThread, 1000);

	  // Close the event and socket
	  shutdown(lpCurrent->hSocket, SD_BOTH);
	  closesocket(lpCurrent->hSocket);
	}

	// Next registration
	lpCurrent = (LPREGISTRATION_INFO) lpCurrent->next;
  }

  // Update the service manager on our closed status
/*  if (!bDebugMode) {
    stat.dwCheckPoint = 0;
    stat.dwCurrentState = SERVICE_STOPPED;
    SetServiceStatus(sshService, &stat);
  } else
  {
    // Remove our handler from the console signal handler list
    SetConsoleCtrlHandler((PHANDLER_ROUTINE) SignalHandler, false);
  }*/
  return 0;
}

// MyTaskBarAddIcon - adds an icon to the taskbar status area. 
// Returns TRUE if successful or FALSE otherwise. 
// hwnd - handle of the window to receive callback messages 
// uID - identifier of the icon 
// hicon - handle of the icon to add 
// lpszTip - ToolTip text 
BOOL MyTaskBarAddIcon(HWND hwnd, UINT uID, HICON hicon, LPSTR lpszTip) 
{ 
    BOOL res; 
    NOTIFYICONDATA tnid; 
 
    tnid.cbSize = sizeof(NOTIFYICONDATA); 
    tnid.hWnd = hwnd; 
    tnid.uID = uID; 
    tnid.uFlags = NIF_MESSAGE | NIF_ICON | NIF_TIP;
    tnid.uCallbackMessage = MYWM_NOTIFYICON; 
    tnid.hIcon = hicon; 
    if (lpszTip) 
        lstrcpyn(tnid.szTip, lpszTip, sizeof(tnid.szTip)); 
    else 
        tnid.szTip[0] = '\0'; 
 
    res = Shell_NotifyIcon(NIM_ADD, &tnid); 
 
    if (hicon) 
        DestroyIcon(hicon); 
 
    return res; 
} 
 

// MyTaskBarDeleteIcon - deletes an icon from the taskbar 
//     status area. 
// Returns TRUE if successful or FALSE otherwise. 
// hwnd - handle of the window that added the icon 
// uID - identifier of the icon to delete 
BOOL MyTaskBarDeleteIcon(HWND hwnd, UINT uID) 
{ 
    BOOL res; 
    NOTIFYICONDATA tnid; 
 
    tnid.cbSize = sizeof(NOTIFYICONDATA); 
    tnid.hWnd = hwnd; 
    tnid.uID = uID; 
         
    res = Shell_NotifyIcon(NIM_DELETE, &tnid); 
    return res; 
} 

LRESULT WINAPI SystemTrayCallbackProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  switch (uMsg)
  {
  case MYWM_NOTIFYICON :
	{
	  UINT uMouseMsg = (UINT) lParam;
	  switch (uMouseMsg) {

	  case WM_RBUTTONUP:
	  case WM_LBUTTONDOWN :
		{
		  /*HMENU hMenu = LoadMenu(ghInstance, MAKEINTRESOURCE(IDR_MENU));
		  DWORD dwButton = (uMouseMsg == WM_RBUTTONUP)? TPM_RIGHTBUTTON: TPM_LEFTBUTTON;
		  POINT pt;
		  GetCursorPos(&pt);
		  if (TrackPopupMenu(hMenu, TPM_RIGHTALIGN | TPM_BOTTOMALIGN | TPM_RETURNCMD | TPM_NONOTIFY | dwButton, pt.x, pt.y, 0, hWnd, NULL) == CM_CLOSE)*/
		  if (bAsking) {
			MessageBeep(MB_ICONASTERISK);
			return 0;
		  }
		  bAsking = true;
		  if (MessageBox(GetTopWindow(NULL), "Shutdown PortProxy?", "PortProxyW95", MB_ICONQUESTION | MB_YESNO | MB_DEFBUTTON2) == IDYES)
		  {
			// Time to close ;-)
			bExit = true;
            PostQuitMessage(0);
		  }
		  bAsking = false;
		}
	  }

	  return 0;
	}
  }
  return DefWindowProc(hWnd, uMsg, wParam, lParam);
}

#define CLASS_NAME "PortProxyW95SystemTrayClass01_version=1.0W95__+0_rev1"
 
// Service primary entry point
int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
  // Start up the sockets
  WSADATA wsaData;
  if (WSAStartup(MAKEWORD(2, 0), &wsaData)) {
	printf("Could not initilize windows sockets!  Error code %lu returned.\n", WSAGetLastError());
	return 0;
  }

  // Get our filename
  char szFileName[65536];
  char *argv[1];
  argv[1] = szFileName;
  GetModuleFileName(NULL, szFileName, 65536);



  WNDCLASS wc;
  wc.style = 0;
  wc.lpfnWndProc = (WNDPROC) SystemTrayCallbackProc;
  wc.cbClsExtra = 0;
  wc.cbWndExtra = 0;
  wc.hInstance = hInstance;
  wc.hIcon = LoadIcon(hInstance, MAKEINTRESOURCE(IDI_ICON));
  wc.hCursor = NULL;
  wc.hbrBackground = NULL;
  wc.lpszMenuName = NULL;
  wc.lpszClassName = CLASS_NAME;
  if (!RegisterClass(&wc)) {
	MessageBox(HWND_DESKTOP, "could not register window class!", "ohdear", MB_OK);
	return 0;
  }
  ghInstance = hInstance;
  bAsking = false;
  HWND hWnd = CreateWindow(CLASS_NAME, "PortProxyW95", WS_POPUP, 0, 0, 0, 0, HWND_DESKTOP, NULL, hInstance, NULL);


  MyTaskBarAddIcon(hWnd, 1, wc.hIcon, "PortProxyW95 Running");

  // Determine what the requested operation is; unfortunately, C++ doesn't have elseif :-(
//  BOOL bServiced = false;
/*  if (IsFlagSet(argv, "-debug"))
  {*/
	// Enter debug (non-service/console) mode
//	bDebugMode = true;
//rintf("PortProxyW95 1.0 ready.");

  // Create a new thread to handle the "service"
  DWORD dwThreadID;
  HANDLE hThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE) ServiceMain, argv, 0, &dwThreadID);
//  ServiceMain(1, argv);

  // Handle our icon's window messages
  MSG msg;
  while (GetMessage(&msg, hWnd, 0, 0))
  {
	TranslateMessage(&msg);
	DispatchMessage(&msg);
  }

  // Cleanup
  MyTaskBarDeleteIcon(hWnd, 1);
  DestroyWindow(hWnd);
  UnregisterClass(CLASS_NAME, hInstance);

  // Wait for the main thread to exit
  bExit = true;
  WaitForSingleObject(hThread, 20000);
  CloseHandle(hThread);
  MessageBox(GetTopWindow(NULL), "PortProxy unloaded.", "PortProxyW95", MB_ICONASTERISK | MB_OK);

//	printf("Exited.\n");
//	bServiced = true;
/*  }
  if (IsFlagSet(argv, "-install") && !bServiced)
  {
	// Install this service with the Control Manager
	SC_HANDLE hSCM = OpenSCManager(NULL, SERVICES_ACTIVE_DATABASE, SC_MANAGER_CONNECT | SC_MANAGER_CREATE_SERVICE | SC_MANAGER_ENUMERATE_SERVICE);
    if (!hSCM) {
	  if (GetLastError() == ERROR_ACCESS_DENIED)
		MessageBox(HWND_DESKTOP, "Access denied while opening the Service Control Manager, installation aborted.", SERVICE_NAME, MB_OK | MB_ICONEXCLAMATION);
	  else
		MessageBox(HWND_DESKTOP, "Severe Windows NT error while opening the Service Control Manager!  Please check your installation.", SERVICE_NAME, MB_OK | MB_ICONSTOP);
	} else
	{
	  // Try to open the service to see if it already exists
	  SC_HANDLE hService = OpenService(hSCM, SERVICE_NAME, 0);
	  if (!hService)
	  {
		// If this service doesn't exist, then add it
		if (GetLastError() == ERROR_SERVICE_DOES_NOT_EXIST) {
		  hService = CreateService(hSCM, SERVICE_NAME, SERVICE_NAME, 0, SERVICE_WIN32_OWN_PROCESS, SERVICE_AUTO_START, SERVICE_ERROR_NORMAL, szFileName, NULL, NULL, "Tcpip\0\0", NULL, NULL);
		  if (!hService)
			MessageBox(HWND_DESKTOP, "Error while creating the service, installation aborted.", SERVICE_NAME, MB_OK | MB_ICONEXCLAMATION);
		  else
		  {
			// Write a success message if we're not supposed to be quiet
			if (!IsFlagSet(argv, "-q"))
			  MessageBox(HWND_DESKTOP, "The service was created and is ready to be started.", SERVICE_NAME, MB_OK | MB_ICONASTERISK);
		  }
		} else {
		  if (GetLastError() == ERROR_ACCESS_DENIED)
		    MessageBox(HWND_DESKTOP, "Access denied while checking for an existing installation of the service.  Have the system administrator uninstall the existing service.", SERVICE_NAME, MB_OK | MB_ICONEXCLAMATION);
		  else
		    MessageBox(HWND_DESKTOP, "Severe Windows NT error while opening the service!  Please check your installation.", SERVICE_NAME, MB_OK | MB_ICONSTOP);
		}
	  } else
		MessageBox(HWND_DESKTOP, "Service PortProxy already exists in the Service Control Manager!  Installation aborted.", SERVICE_NAME, MB_OK | MB_ICONEXCLAMATION);

	  // Clean up the handles
	  if (hService)
	    CloseServiceHandle(hService);
	  CloseServiceHandle(hSCM);
	}
	bServiced = true;
  }
  if (IsFlagSet(argv, "-uninstall") && !bServiced)
  {
	// Remove this service from the Control Manager
	SC_HANDLE hSCM = OpenSCManager(NULL, SERVICES_ACTIVE_DATABASE, SC_MANAGER_CONNECT | SC_MANAGER_ENUMERATE_SERVICE);
    if (!hSCM) {
	  if (GetLastError() == ERROR_ACCESS_DENIED)
		MessageBox(HWND_DESKTOP, "Access denied while opening the Service Control Manager, installation aborted.", SERVICE_NAME, MB_OK | MB_ICONEXCLAMATION);
	  else
		MessageBox(HWND_DESKTOP, "Severe Windows NT error while opening the Service Control Manager!  Please check your installation.", SERVICE_NAME, MB_OK | MB_ICONSTOP);
	} else
	{
	  // Try to open the service to see if it already exists
	  SC_HANDLE hService = OpenService(hSCM, SERVICE_NAME, DELETE);
	  if (!hService) {
		if (GetLastError() == ERROR_ACCESS_DENIED)
		  MessageBox(HWND_DESKTOP, "Access denied while deleting the service, uninstallation aborted.", SERVICE_NAME, MB_OK | MB_ICONEXCLAMATION);
		else {
    	  if (GetLastError() == ERROR_SERVICE_DOES_NOT_EXIST)
			MessageBox(HWND_DESKTOP, "Service does not exist, uninstallation aborted.", SERVICE_NAME, MB_OK | MB_ICONEXCLAMATION);
		  else
			MessageBox(HWND_DESKTOP, "Severe Windows NT error while opening the service!  Please check your installation.", SERVICE_NAME, MB_OK | MB_ICONSTOP);
		}
	  } else
	  {
		// Delete the service
		if (!DeleteService(hService)) {
		  if (GetLastError() == ERROR_SERVICE_MARKED_FOR_DELETE)
			MessageBox(HWND_DESKTOP, "Service already marked for deletion, uninstallation aborted.", SERVICE_NAME, MB_OK | MB_ICONEXCLAMATION);
		  else
		    MessageBox(HWND_DESKTOP, "Severe Windows NT error while deleting the service!  Please check your installation.", SERVICE_NAME, MB_OK | MB_ICONSTOP);
		} else {
		  // Write a success message if we're not supposed to be quiet
		  if (!IsFlagSet(argv, "-q"))
		    MessageBox(HWND_DESKTOP, "The service was stopped (if running) and deleted from this computer.", SERVICE_NAME, MB_OK | MB_ICONASTERISK);
		}
	  }

	  // Clean up the handles
	  if (hService)
	    CloseServiceHandle(hService);
	  CloseServiceHandle(hSCM);
	}
	bServiced = true;
  }
  if (!bServiced && !(IsFlagSet(argv, "--help") || IsFlagSet(argv, "/?")))
  {
    // Start up the service
    SERVICE_TABLE_ENTRY ste;
    ste.lpServiceProc = (LPSERVICE_MAIN_FUNCTION) ServiceMain;
    ste.lpServiceName = SERVICE_NAME;
	bDebugMode = false;
    if (!StartServiceCtrlDispatcher(&ste)) {
	  printf("Call to StartServiceCtrlDispatcher for %s failed!  Error code %lu returned.\n\n", SERVICE_NAME, GetLastError());
	} else
	  bServiced = true;
  }
  if (!bServiced)
  {
	// Display the command line options
	printf("Command line options:\n");
	printf("%s [[-install [-q]] | [-uninstall [-q]] | [-debug] | [[--help] | [/?]]]\n\n", szFileName);
	printf("-install      Install the service\n");
	printf("-uninstall    Uninstall the service\n");
	printf("-q            Use with -install or -uninstall for quiet operations\n");
	printf("-debug        Run %s as a program instead of a service\n", SERVICE_NAME);
	printf("--help or /?  Display this screen");
  }*/

  // Shut down the sockets
  WSACleanup();
  return 0;
}
